from .version import __version__
from .row_welch import *
from .reference_families import *
from .post_hoc_bounds import *
from .lambda_calibration import *


__all__ = [
    'get_summary_stats',
    'suff_welch_test',
    'row_welch_tests',
    'inverse_linear_template',
    'inverse_beta_template',
    'linear_template',
    'beta_template',
    'get_permuted_p_values',
    'get_permuted_p_values_one_sample',
    'get_pivotal_stats',
    'max_fp',
    'min_tp',
    'curve_max_fp',
]



import numpy as np
from scipy.stats import norm


# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# POST HOC BOUNDS
# R source code: https://github.com/pneuvial/sanssouci/
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def max_fp(p_values, thresholds):
    """
    Upper bound for the number of false discoveries in a selection

    Parameters
    ----------

    p_values : 1D numpy.array
        A 1D numpy array of p-values for the selected items
    thresholds : 1D numpy.array
        A 1D numpy array of non-decreasing k-FWER-controlling thresholds

    Returns
    -------

    scalar :
        A post hoc upper bound on the number of false discoveries in the
        selection

    See Also
    --------

    sanssouci.min_tp, sanssouci.curve_max_fp

    References
    ----------

    .. [1] Blanchard, G., Neuvial, P., & Roquain, E. (2020). Post hoc
        confidence bounds on false positives using reference families.
        Annals of Statistics, 48(3), 1281-1303.
    """

    # make sure that thresholds is sorted
    if np.linalg.norm(thresholds - np.sort(thresholds)) \
       > 0.0001:
        thresholds = np.sort(thresholds)
        print("The input 'thresholds' was not sorted -> this is done now")

    # do the job
    subset_size = p_values.shape[0]
    template_size = thresholds.shape[0]
    size = np.min([subset_size, template_size])

    if size < 1:
        return 0

    seq_k = np.arange(size)

    # k-FWER control for k>subset_size is useless
    # (will yield bound > subset_size)
    thresholds = thresholds[seq_k]

    card = np.zeros(thresholds.shape[0])
    for i in range(thresholds.shape[0]):
        card[i] = np.sum(p_values > thresholds[i])

    return np.min([subset_size, (card + seq_k).min()])


def min_tp(p_values, thresholds):
    """
    Lower bound for the number of true discoveries in a selection

    Parameters
    -----------

    p_values : 1D numpy.array
        A 1D numpy array of p-values for the selected items
    thresholds : 1D numpy.array
        A 1D numpy array of non-decreasing k-FWER-controlling thresholds

    Returns
    -------

    scalar :
        A Lower bound on the number of true discoveries in the selection

    References
    ----------

    .. [1] Blanchard, G., Neuvial, P., & Roquain, E. (2020). Post hoc
        confidence bounds on false positives using reference families.
        Annals of Statistics, 48(3), 1281-1303.
    """

    return p_values.shape[0] - max_fp(p_values, thresholds)


def min_tdp(p_values, thresholds):
    """Lower bound for the proportion of true discoveries in a selection
    Lower bound for the proportion of true discoveries in a selection

    Parameters
    -----------

    p_values : 1D numpy.array
        A 1D numpy array of p-values for the selected items
    thresholds : 1D numpy.array
        A 1D numpy array of non-decreasing k-FWER-controlling thresholds

    Returns
    -------

    scalar :
        A Lower bound on the proportion of true discoveries in the selection

    References
    ----------

    .. [1] Blanchard, G., Neuvial, P., & Roquain, E. (2020). Post hoc
        confidence bounds on false positives using reference families.
        Annals of Statistics, 48(3), 1281-1303.
    """
    if len(p_values) == 0:
        return 0
    else:
        return min_tp(p_values, thresholds) / len(p_values)


def curve_max_fp(p_values, thresholds):
    """
    Upper bound for the number of false discoveries among most
    significant items.

    Parameters
    ----------

    p_values : 1D numpy.array
        A 1D numpy array containing all p-values,sorted non-decreasingly
    thresholds : 1D numpy.array
        A 1D numpy array  of K JER-controlling thresholds,
        sorted non-decreasingly

    Returns
    -------

    numpy.array :
        A vector of size p giving an joint upper confidence bound on the
        number of false discoveries among the k most significant items for
        all k in \{1,\ldots,m\}

    References
    ----------

    .. [1] Blanchard, G., Neuvial, P., & Roquain, E. (2020). Post hoc
        confidence bounds on false positives using reference families.
        Annals of Statistics, 48(3), 1281-1303.
    """

    #  make sure that p_values and thresholds are sorted
    if np.linalg.norm(p_values - p_values[np.argsort(p_values)]) > 0.0001:
        p_values = p_values[np.argsort(p_values)]
        print("The input p-values were not sorted -> this is done now")

    if np.linalg.norm(thresholds - thresholds[np.argsort(thresholds)]) \
       > 0.0001:
        thresholds = thresholds[np.argsort(thresholds)]
        print("The input 'thresholds' were not sorted -> this is done now")

    # do the job
    p = p_values.shape[0]
    k_max = thresholds.shape[0]

    if k_max < p:
        thresholds = np.concatenate((thresholds, thresholds[-1] *
                                    np.ones(p - k_max)))
        k_max = thresholds.shape[0]

    K = np.ones(p) * (k_max)
    # K[i] = number of k/ T[i] <= s[k] = BB in 'Mein2006'
    Z = np.ones(k_max) * (p)
    # Z[k] = number of i/ T[i] >  s[k] = cardinal of R_k
    # 'K' and 'Z' are initialized to their largest possible value,
    #       ie 'p' and 'k_max', respectively

    kk = 0
    ii = 0

    while (kk < k_max) and (ii < p):
        if thresholds[kk] >= p_values[ii]:
            K[ii] = kk
            ii += 1
        else:
            Z[kk] = ii
            kk += 1

    max_fp_ = np.zeros(p)
    indices = np.where(K > 0)[0]
    A = Z - np.arange(0, k_max)

    K_ww = K[K > 0].astype(np.int)
    cummax_A = A.copy()
    for i in range(1, cummax_A.shape[0]):
        cummax_A[i] = np.max([cummax_A[i - 1], cummax_A[i]])

    cA = cummax_A[K_ww - 1]  # cA[i] = max_{k<K[i]} A[k]

    max_fp_[K > 0] = np.min(
        np.concatenate(((indices + 1 - cA).reshape(1, -1),
                        (K[K > 0]).reshape(1, -1)),
                       axis=0), axis=0)

    return max_fp_


def curve_min_tdp(p_values, thresholds):
    """
    Lower TDP bounds among most
    significant items.

    Parameters
    ----------

    p_values : 1D numpy.array
        A 1D numpy array containing all p-values,sorted non-decreasingly
    thresholds : 1D numpy.array
        A 1D numpy array  of K JER-controlling thresholds,
        sorted non-decreasingly

    Returns
    -------

    numpy.array :
        A vector of size p giving an joint lower confidence bound on the
        true discovery proportion among the k most significant items for
        all k in \{1,\ldots,m\}

    References
    ----------

    .. [1] Blanchard, G., Neuvial, P., & Roquain, E. (2020). Post hoc
        confidence bounds on false positives using reference families.
        Annals of Statistics, 48(3), 1281-1303.
    """
    p = p_values.shape[0]
    range = np.arange(1, p + 1)  # use this to shorten following line
    return (range - curve_max_fp(p_values, thresholds)) / range


def find_largest_region(p_values, thresholds, tdp, masker=None):
    """
    Find largest FDP controlling region.

    Parameters
    ----------

    p_values : 1D numpy.array
        A 1D numpy array containing all p-values,sorted non-decreasingly
    thresholds : 1D numpy.array
        A 1D numpy array  of K JER-controlling thresholds,
        sorted non-decreasingly
    tdp : float
        True discovery proportion
    masker: NiftiMasker
        masker used on current data

    Returns
    -------

    z_unmasked_cal : nifti image of z_values of the FDP controlling region
    region_size : size of TDP controlling region

    """
    z_map_ = norm.isf(p_values)

    res = curve_min_tdp(p_values, thresholds)
    region_size = len(res[res > tdp])
    pval_cutoff = sorted(p_values)[region_size - 1]
    z_cutoff = norm.isf(pval_cutoff)

    if masker is not None:
        z_to_plot = np.copy(z_map_)
        z_to_plot[z_to_plot < z_cutoff] = 0
        z_unmasked_cal = masker.inverse_transform(z_to_plot)
        return z_unmasked_cal, region_size

    return region_size


def _compute_hommel_value(z_vals, alpha):
    """Compute the All-Resolution Inference hommel-value
    Function taken from nilearn.glm
    """
    if alpha < 0 or alpha > 1:
        raise ValueError('alpha should be between 0 and 1')
    z_vals_ = - np.sort(- z_vals)
    p_vals = norm.sf(z_vals_)
    n_tests = len(p_vals)

    if len(p_vals) == 1:
        return p_vals[0] > alpha
    if p_vals[0] > alpha:
        return n_tests
    slopes = (alpha - p_vals[: - 1]) / np.arange(n_tests, 1, -1)
    slope = np.max(slopes)
    hommel_value = np.trunc(n_tests + (alpha - slope * n_tests) / slope)
    return np.minimum(hommel_value, n_tests)



import numpy as np
from scipy import stats

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROW WELCH TESTS
# R source code: https://github.com/pneuvial/sanssouci/
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


def get_summary_stats(X, labels):
    """
    Convert a matrix of observations labelled into categories into summary
    statistics for each category

    The following statistics are calculated: sums, sums of squares, means,
    standard deviations, sample sizes

    Parameters
    ----------

    X : array-like of shape (n, p)
        Numpy array matrix whose columns correspond to the p variables
        and rows to the n observations
    labels :  array-like of shape (n, )
        A numpy array of size n representing the label of each
        observation, in {0, 1}

    Returns
    -------

    test : dict of float arrays of shape (p,)
        A dictionary containing the above-described summary statistics for
        each category
    """

    labels_set = set(labels)

    res = {}

    for lab in labels_set:
        X_lab = X[labels == lab]
        sum_lab = np.sum(X_lab, axis=0)
        sum_squares_lab = np.sum(X_lab * X_lab, axis=0)
        n_lab = np.sum(labels == lab)
        mean_lab = sum_lab / n_lab
        std_lab = np.sqrt((sum_squares_lab -
                           ((sum_lab * sum_lab) / n_lab)) / (n_lab - 1))
        res[lab] = {"sum": sum_lab, "sum2": sum_squares_lab, "n": n_lab,
                    "mean": mean_lab, "sd": std_lab}

    return res


def suff_welch_test(mean_x, mean_y, std_x, std_y, n_x, n_y):
    """
    Welch test from sufficient statistics

    Parameters
    ----------

    mean_x : array-like
        A numeric value or vector, the sample average for condition "x"
    mean_y : array-like
        A numeric value or vector of the same length as 'mean_x', the sample
        average for condition "y"
    std_x : array-like
        A numeric value or vector of the same length as 'mean_x', the standard
        deviation for condition "x"
    std_y : array-like
        A numeric value or vector of the same length as 'mean_x', the standard
        deviation for condition "y"
    n_x : array-like
        A numeric value or vector of the same length as 'mean_x', the sample
        size for condition "x"
    n_y : array-like
        A numeric value or vector of the same length as 'mean_x', the sample
        size for condition "y"

    Returns
    -------

    test : dict of float arrays of shape (p,) with
        statistic: the value of the t-statistic
        parameter:  the degrees of freedom for the t-statistic
        p_value: the p-value for the test

    Notes
    -----

    Note that the alternative hypothesis is "two.sided". It could be extended
    to "greater" or "less" as in the original R code.
    """

    # pre-computations: squared standard error of the mean (sem)
    squared_sem_x = (std_x * std_x) / n_x
    squared_sem_y = (std_y * std_y) / n_y
    squared_sem = squared_sem_x + squared_sem_y

    # test statistic
    stat = (mean_x - mean_y) / np.sqrt(squared_sem)

    # approximate degrees of freedom (Welch-Satterthwaite)
    df = squared_sem * squared_sem /\
        ((squared_sem_x * squared_sem_x / (n_x - 1)) +
         ((squared_sem_y * squared_sem_y) / (n_y - 1)))

    # two-sided p-value
    p_value = 2 * (1 - stats.t.cdf(np.abs(stat), df=df))

    return {"statistic": stat, "parameter": df, "p_value": p_value}


def row_welch_tests(X, labels):
    """
    Welch t-tests for each column of a matrix, intended to be speed efficient

    Note that the alternative hypothesis is "two.sided". It could be extended
    to "greater" or "less" as in the original R code.

    Parameters
    ----------

    X : array-like of shape (n, p)
        Numpy array matrix whose columns correspond to the p variables
         and rows to the n observations
    labels :  array-like of shape (n, )
        A numpy array of size n representing the category of each
        observation, in {0, 1}

    Returns
    -------

    test : dict of float arrays of shape (p,) with
        statistic: the value of the t-statistic
        parameter:  the degrees of freedom for the t-statistic
        p_value: the p-value for the test
        meanDiff: the difference between means

    Notes
    -----

    Note that the alternative hypothesis is "two.sided". It could be extended
    to "greater" or "less" as in the original R code.

    References
    ----------

    ..[1] B. L. Welch (1951), On the comparison of several mean values:
        an alternative approach. Biometrika, 38, 330-336

    """

    summary_stats = get_summary_stats(X, labels)

    Y = summary_stats[0]
    X = summary_stats[1]

    welch = suff_welch_test(X["mean"], Y["mean"], X["sd"],
                            Y["sd"], X["n"], Y["n"])
    welch["meanDiff"] = X["mean"] - Y["mean"]

    return welch



import numpy as np
from scipy.stats import beta


# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# REFERENCE FAMILIES
# R source code: https://github.com/pneuvial/sanssouci/
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def inverse_linear_template(y, k, m):
    """
    Parameters
    ----------
    y : array of floats of shape (B, )
        values to apply template to
    k : int
        index of rejection set in linear template
    m : int
        number of hypotheses

    Returns
    -------

    estimated_template : array of same shape as y

    """

    return y * m / k


def linear_template(alpha, k, m):
    """

    Parameters
    ----------

    alpha : float
        confidence level in [0, 1]
    k : int
        number of rejection sets in linear template
    m : int
        number of hypotheses

    Returns
    -------
    template : array of shape (k,)
    """
    return alpha * np.arange(1, k + 1) / m


def beta_template(alpha, k, m):
    """

    Parameters
    ----------

    alpha : float
        confidence level in [0, 1]
    k : int
        number of rejection sets in Beta template
    m : int
        number of hypotheses

    Returns
    -------
    template : array of shape (k,)
    """
    return beta.ppf(alpha, np.arange(1, k + 1),
                    np.array([m + 1] * k) - np.arange(1, k + 1))


def inverse_beta_template(y, k, m):
    """
    Parameters
    ----------
    y : array of floats of shape (B, )
        values to apply template to
    k : int
        index of rejection set in beta template
    m : int
        number of hypotheses

    Returns
    -------

    estimated_template : array of same shape as y

    """
    return beta.cdf(y, k, m + 1 - k)


def _regularised_inverse_beta_template(y, k, m, reg=100):
    """
    Parameters
    ----------
    y : array of floats of shape (B, )
        values to apply template to
    k : int
        index of rejection set in beta template
    m : int
        number of hypotheses
    reg : float
        scaling factor for beta law

    Returns
    -------

    estimated_template : array of same shape as y

    """
    return beta.cdf(y, k / reg, (m + 1 - k) / reg)


def _regularised_beta_template(alpha, k, m, reg=100):
    """
    Beta template with scaled parameters

    Parameters
    ----------

    alpha : float
        confidence level in [0, 1]
    k : int
        number of rejection sets in Beta template
    m : int
        number of hypotheses
    reg : float
        scaling factor for beta law

    Returns
    -------
    template : array of shape (k,)
    """
    return beta.ppf(alpha, (np.arange(1, k + 1)) / reg,
                    (np.array([m + 1] * k) - np.arange(1, k + 1)) / reg)


def quadratic_template(alpha, k, m):
    """
    Linear template with an additional quadratic term

    Parameters
    ----------

    alpha : float
        confidence level in [0, 1]
    k : int
        number of rejection sets in linear template
    m : int
        number of hypotheses

    Returns
    -------
    template : array of shape (k,)
    """

    return alpha * np.arange(1, k + 1) / m + \
        (1 - alpha) * np.square(np.arange(1, k + 1) / m)


def inverse_quadratic_template(y, k, m):
    """
    Parameters
    ----------
    y : array of floats
        values to apply template to
    k : int
        index of rejection set in linear template
    m : int
        number of hypotheses

    Returns
    -------

    estimated_template : array of same shape as y

    """
    if k == m:
        return np.ones(len(y))
    else:
        return (np.array(y) * m ** 2 - k ** 2) / (k * (m - k))



import numpy as np
from scipy import stats

from .row_welch import row_welch_tests
from .reference_families import inverse_linear_template, inverse_beta_template

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# LAMBDA-CALIBRATION
# R source code: https://github.com/pneuvial/sanssouci/
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


def get_permuted_p_values(X, labels, B=100, row_test_fun=stats.ttest_ind):
    """
    Get permutation p-values: Get a matrix of p-values under the null
    hypothesis obtained by repeated permutation of class labels.

    Parameters
    ----------

    X : array-like of shape (n,p)
        numpy array of size [n,p], containing n observations of p variables
        (hypotheses)
    labels : array-like of shape (n,)
        numpy array of size [n], containing n values in {0, 1}, each of them
        specifying the column indices of the first and the second sample.
    B : int
        number of permutations to be performed (default=100)
    row_test_fun : function
        testing function with the same I/O as 'stats.ttest_ind' (default).
        Specifically, must have two lists as inputs (or 1d np.arrays) for
        thecompared data, and the resulting pvalue must be accessed in
        '[test].pvalue' Eligible functions are for instance "stats.ks_2samp",
        "stats.bartlett", "stats.ranksums", "stats.kruskal"

    Returns
    -------

    pval0 : array-like of shape (B, p)
        A numpy array of size [B,p], whose entry i,j corresponds to
        p_{(j)}(g_i.X) with notation of the AoS 2020 paper cited below
        (section 4.5) [1]_

    References
    ----------

    .. [1] Blanchard, G., Neuvial, P., & Roquain, E. (2020). Post hoc
        confidence bounds on false positives using reference families.
        Annals of Statistics, 48(3), 1281-1303.
    """

    # Init
    n, p = X.shape

    # Step 1: calculate $p$-values for B permutations of the class assignments

    # 1.1: Intialise all vectors and matrices
    shuffled_labels = labels.copy()

    all_shuffled_labels = np.zeros((B, n))
    for bb in range(B):
        np.random.shuffle(shuffled_labels)
        all_shuffled_labels[bb] = shuffled_labels

    # 1.2: calculate the p-values
    pval0 = np.zeros([B, p])

    if row_test_fun == row_welch_tests:  # row Welch Tests (parallelized)
        for b in range(B):
            permuted_test_result = row_welch_tests(X, all_shuffled_labels[b])
            pval0[b] = permuted_test_result['p_value']
    else:                     # standard scipy tests
        for b in range(B):
            s0 = np.where(all_shuffled_labels[b] == 0)[0]
            s1 = np.where(all_shuffled_labels[b] == 1)[0]

            for ii in range(p):
                rwt = row_test_fun(X[s0, ii], X[s1, ii])
                # Welch test with scipy -> rwt=stats.ttest_ind(X[s0, ii],
                # X[s1, ii], equal_var=False)

                pval0[b, ii] = rwt.pvalue

    # Step 2: sort each column
    pval0 = np.sort(pval0, axis=1)

    return pval0


def get_permuted_p_values_one_sample(X, B=100, seed=None):
    """
    Get permutation p-values: Get a matrix of p-values under the null
    hypothesis obtained by sign-flipping (one-sample test).

    Parameters
    ----------

    X : array-like of shape (n,p)
        numpy array of size [n,p], containing n observations of p variables
        (hypotheses)
    B : int
        number of sign-flippings to be performed (default=100)

    Returns
    -------

    pval0 : array-like of shape (B, p)
        A numpy array of size [B,p], whose rows are sorted increasingly.
        The entry i,j corresponds to p_{(j)}(g_i.X) with notation of [1]
        (section 4.5)_

    References
    ----------

    .. [1] Blanchard, G., Neuvial, P., & Roquain, E. (2020). Post hoc
        confidence bounds on false positives using reference families.
        Annals of Statistics, 48(3), 1281-1303.
    """

    np.random.seed(seed)

    # Init
    n, p = X.shape

    # intialise p-values
    pval0 = np.zeros((B, p))

    for b in range(B):
        X_flipped = (X.T * (2 * np.random.randint(-1, 1, size=n) + 1)).T
        _, pval0[b] = stats.ttest_1samp(X_flipped, 0)

    # Sort each column
    pval0 = np.sort(pval0, axis=1)

    return pval0


def get_pivotal_stats(p0, inverse_template=inverse_linear_template, K=-1):
    """Get pivotal statistic

    Parameters
    ----------

    p0 :  array-like of shape (B, p)
        A numpy array of size [B,p] of null p-values obtained from
        B permutations for p hypotheses.
    inverse_template : function
        A function with the same I/O as inverse_template_linear
    K :  int
        For JER control over 1:K, i.e. joint control of all k-FWER, k<= K.
        Automatically set to p if its input value is < 0.

    Returns
    -------

    array-like of shape (B,)
        A numpy array of of size [B]  containing the pivotal statitics, whose
        j-th entry corresponds to \psi(g_j.X) with notation of the AoS 2020
        paper cited below (section 4.5) [1]_

    References
    ----------

    .. [1] Blanchard, G., Neuvial, P., & Roquain, E. (2020). Post hoc
        confidence bounds on false positives using reference families.
        Annals of Statistics, 48(3), 1281-1303.
    """
    # Sort permuted p-values
    p0 = np.sort(p0, axis=1)

    # Step 3: apply template function
    # For each feature p, compare sorted permuted p-values to template
    B, p = p0.shape
    tk_inv_all = np.array([inverse_template(p0[:, i], i + 1, p)
                           for i in range(p)]).T

    if K < 0:
        K = tk_inv_all.shape[1]  # tkInv_all.shape[1] is equal to p

    # Step 4: report min for each row
    pivotal_stats = np.min(tk_inv_all[:, :K], axis=1)

    return pivotal_stats


def estimate_jer(template, pval0, k_max):

    """
    Compute empirical JER for a given template and permuted p-values
    """

    B, p = pval0.shape
    id_ranks = np.tile(np.arange(0, p), (B, 1))

    cutoffs = np.searchsorted(template, pval0)

    signs = np.sign(id_ranks - cutoffs)
    sgn_trunc = signs[:, :k_max]
    JER = np.sum([np.any(sgn_trunc[perm] >= 0) for perm in range(B)]) / B

    return JER


def calibrate_jer(alpha, learned_templates, pval0, k_max, min_dist=3):

    """
    For a given risk level, calibrate the method on learned templates by
    dichotomy. This is equivalent to calibrating using pivotal stats but does
    not require the availability of a closed form inverse template.

    Parameters
    ----------

    alpha : float
        confidence level in [0, 1]
    learned_templates : array of shape (B', p)
        learned templates for B' permutations and p voxels
    pval0 :  array of shape (B, p)
        permuted p-values
    k_max : int
        template size
    min_dist : int
        minimum distance to stop iterating dichotomy

    Returns
    -------

    int : index of template chosen by calibration

    """

    B, p = learned_templates.shape
    low, high = 0, B - 1
    while high - low > min_dist:
        mid = int((high + low) / 2)
        lw = estimate_jer(learned_templates[low], pval0, k_max) - alpha
        md = estimate_jer(learned_templates[mid], pval0, k_max) - alpha
        if md == 0:
            return mid
        if lw * md < 0:
            high = mid
        else:
            low = mid
    return low



from setuptools import setup, find_packages

with open("README.md", "r") as readme_file:
    readme = readme_file.read()

requirements = ["numpy>=1.15.0", "scipy>=1.0.0"]

setup(
    name="sanssouci",
    version="0.1.2",
    author="Laurent Risser and Pierre Neuvial",
    author_email="pierre.neuvial@math.univ-toulouse.fr",
    description="Post hoc inference via multiple testing",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/sanssouci.python/",
    packages=find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
)
